#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>

#include <zephyr/drivers/uart.h>
#include <zephyr/kernel.h>
#include <segger_rtt.h>

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// UART Hello World
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
#define UART0 DT_NODELABEL(uart0)
const struct device *const uart_dev = DEVICE_DT_GET(UART0);
struct k_sem uart_rxdrdy;
#define RX_BUFLEN (256)

static volatile char rx_buffer[RX_BUFLEN];
void ASSERT(int x)
{
    if (x)
        return;

    printk("ASSERTION FAILED\n");
    while(1);
}

void uart_callback(const struct device *dev, struct uart_event *evt, void *user_data)
{
    printk("UART Callback %d %8p\n", evt->type, evt->data.rx.buf);
    switch (evt->type)
    {
        case UART_RX_RDY:
            printk("UART RX Ready - notify receiving function\n");
            k_sem_give(&uart_rxdrdy);
            break;
        case UART_RX_BUF_REQUEST:
            printk("UART RX Buf request-ignoring\n");
            break;
        case UART_RX_BUF_RELEASED:
            if (evt->data.rx.len == 0) {
                printk("UART RX Buf %p released, no data\n", evt->data.rx.buf);
            }
            else{
                printk("UART RX Buf released\n");
            }
            break;
        case UART_RX_DISABLED:
            printk("UART RX Disabled\n");
            break;
        case UART_RX_STOPPED:
            printk("UART RX Stopped\n");
            break;
        default:
            printk("UNhandled UART event: %d\n", evt->type);
            break;
    }
}

uint8_t *uart_rx(void)
{
    /* This is run in the context of the main thread */

    /*  We are not expecting to need more than one buffer at a time as we send in a while CW instruction and
    then make a response. The inputs are all well controlled.    */
    for (int j = 0; j < RX_BUFLEN; j++)
    {
        rx_buffer[j] = 0;
    }
    uint8_t *buf = (uint8_t *)rx_buffer;
    printk("UARTRX..  setting up for data into buffer %p\n", buf);
    int ret = uart_rx_enable(uart_dev, buf, RX_BUFLEN, 30*1000*1000);
    ASSERT (ret >= 0);
    int sem_taken = 0;
    int i = 0;
    do{
        printk("%d UARTRX SEM..  sem take\n", i);
        sem_taken = k_sem_take(&uart_rxdrdy, K_MSEC(10000));
        switch (sem_taken)
        {
            case 0:
                printk("%d UARTRX SEM..  sem taken %d\n",i, sem_taken);
                break;
            case -EAGAIN:
                printk("%d UARTRX SEM..  sem timeout %d\n",i, sem_taken);
                break;
            case -EBUSY:
                printk("%d UARTRX SEM..  sem busy %d\n",i, sem_taken);
                break;
        }
        i++;
    } while (sem_taken != 0);

    return buf;
}

void init_uart(void)
{
    k_sem_init(&uart_rxdrdy, 0, 1);
    printk("UART.. device %p\n", uart_dev);
    ASSERT (uart_dev != NULL);

    printk("UART.. Buffer: %p\n", rx_buffer);
    int ret = uart_callback_set(uart_dev, uart_callback, NULL);
    ASSERT (ret == 0);
}


void main(void)
{
	printk("Hello World! %s\n", CONFIG_BOARD);
    init_uart();
	while(1)
	{
		printk("main..  waiting for data\n");
		char *buf = uart_rx();
		printk("main..  have received '%s'\n", buf);
	}
}
